@@ -3,7 +3,8 @@ import threading
 import itertools
 import concurrent.futures
 
-from collections import OrderedDict, namedtuple
+from collections import namedtuple
+from typing import List, Optional
 
 from math import isnan
 
@@ -12,7 +13,7 @@ from scipy.sparse import issparse
 
 from AnyQt.QtWidgets import (
     QTableView, QHeaderView, QAbstractButton, QApplication, QStyleOptionHeader,
-    QStyle, QStylePainter, QStyledItemDelegate
+    QStyle, QStylePainter
 )
 from AnyQt.QtGui import QColor, QClipboard
 from AnyQt.QtCore import (
@@ -30,13 +31,14 @@ from Orange.statistics import basic_stats
 
 from Orange.widgets import gui
 from Orange.widgets.settings import Setting
+from Orange.widgets.utils.itemdelegates import TableDataDelegate
 from Orange.widgets.utils.itemselectionmodel import (
     BlockSelectionModel, ranges, selection_blocks
 )
 from Orange.widgets.utils.tableview import TableView, \
     table_selection_to_mime_data
 from Orange.widgets.utils.widgetpreview import WidgetPreview
-from Orange.widgets.widget import OWWidget, Input, Output
+from Orange.widgets.widget import OWWidget, MultiInput, Output
 from Orange.widgets.utils import datacaching
 from Orange.widgets.utils.annotated_data import (create_annotated_table,
                                                  ANNOTATED_DATA_SIGNAL_NAME)
@@ -170,6 +172,10 @@ class DataTableView(gui.HScrollStepMixin, TableView):
     input_slot: TableSlot
 
 
+class TableBarItemDelegate(gui.TableBarItem, TableDataDelegate):
+    pass
+
+
 class OWDataTable(OWWidget):
     name = "Data Table"
     description = "View the dataset in a spreadsheet."
@@ -178,7 +184,7 @@ class OWDataTable(OWWidget):
     keywords = []
 
     class Inputs:
-        data = Input("Data", Table, multiple=True)
+        data = MultiInput("Data", Table, auto_summary=False, filter_none=True)
 
     class Outputs:
         selected_data = Output("Selected Data", Table, default=True)
@@ -200,9 +206,7 @@ class OWDataTable(OWWidget):
 
     def __init__(self):
         super().__init__()
-
-        self._inputs = OrderedDict()
-
+        self._inputs: List[TableSlot] = []
         self.__pending_selected_rows = self.selected_rows
         self.selected_rows = None
         self.__pending_selected_cols = self.selected_cols
@@ -213,7 +217,6 @@ class OWDataTable(OWWidget):
         info_box = gui.vBox(self.controlArea, "Info")
         self.info_text = gui.widgetLabel(info_box)
         self._set_input_summary(None)
-        self._set_output_summary(None)
 
         box = gui.vBox(self.controlArea, "Variables")
         self.c_show_attribute_labels = gui.checkBox(
@@ -252,78 +255,90 @@ class OWDataTable(OWWidget):
     def sizeHint():
         return QSize(800, 500)
 
+    def _create_table_view(self):
+        view = DataTableView()
+        view.setSortingEnabled(True)
+        view.setItemDelegate(TableDataDelegate(view))
+
+        if self.select_rows:
+            view.setSelectionBehavior(QTableView.SelectRows)
+
+        header = view.horizontalHeader()
+        header.setSectionsMovable(True)
+        header.setSectionsClickable(True)
+        header.setSortIndicatorShown(True)
+        header.setSortIndicator(-1, Qt.AscendingOrder)
+
+        # QHeaderView does not 'reset' the model sort column,
+        # because there is no guaranty (requirement) that the
+        # models understand the -1 sort column.
+        def sort_reset(index, order):
+            if view.model() is not None and index == -1:
+                view.model().sort(index, order)
+        header.sortIndicatorChanged.connect(sort_reset)
+        return view
+
     @Inputs.data
-    def set_dataset(self, data, tid=None):
+    def set_dataset(self, index: int, data: Table):
         """Set the input dataset."""
-        if data is not None:
-            datasetname = getattr(data, "name", "Data")
-            if tid in self._inputs:
-                # update existing input slot
-                slot = self._inputs[tid]
-                view = slot.view
-                # reset the (header) view state.
-                view.setModel(None)
-                view.horizontalHeader().setSortIndicator(-1, Qt.AscendingOrder)
-                assert self.tabs.indexOf(view) != -1
-                self.tabs.setTabText(self.tabs.indexOf(view), datasetname)
-            else:
-                view = DataTableView()
-                view.setSortingEnabled(True)
+        datasetname = getattr(data, "name", "Data")
+        slot = self._inputs[index]
+        view = slot.view
+        # reset the (header) view state.
+        view.setModel(None)
+        view.horizontalHeader().setSortIndicator(-1, Qt.AscendingOrder)
+        assert self.tabs.indexOf(view) != -1
+        self.tabs.setTabText(self.tabs.indexOf(view), datasetname)
+        view.dataset = data
+        slot = TableSlot(index, data, table_summary(data), view)
+        view.input_slot = slot
+        self._inputs[index] = slot
+        self._setup_table_view(view, data)
+        self.tabs.setCurrentWidget(view)
 
-                if self.select_rows:
-                    view.setSelectionBehavior(QTableView.SelectRows)
+    @Inputs.data.insert
+    def insert_dataset(self, index: int, data: Table):
+        datasetname = getattr(data, "name", "Data")
+        view = self._create_table_view()
+        slot = TableSlot(None, data, table_summary(data), view)
+        view.dataset = data
+        view.input_slot = slot
+        self._inputs.insert(index, slot)
+        self.tabs.insertTab(index, view, datasetname)
+        self._setup_table_view(view, data)
+        self.tabs.setCurrentWidget(view)
 
-                header = view.horizontalHeader()
-                header.setSectionsMovable(True)
-                header.setSectionsClickable(True)
-                header.setSortIndicatorShown(True)
-                header.setSortIndicator(-1, Qt.AscendingOrder)
+    @Inputs.data.remove
+    def remove_dataset(self, index):
+        slot = self._inputs.pop(index)
+        view = slot.view
+        self.tabs.removeTab(self.tabs.indexOf(view))
+        view.setModel(None)
+        view.hide()
+        view.deleteLater()
 
-                # QHeaderView does not 'reset' the model sort column,
-                # because there is no guaranty (requirement) that the
-                # models understand the -1 sort column.
-                def sort_reset(index, order):
-                    if view.model() is not None and index == -1:
-                        view.model().sort(index, order)
-
-                header.sortIndicatorChanged.connect(sort_reset)
-                self.tabs.addTab(view, datasetname)
-
-            view.dataset = data
-            self.tabs.setCurrentWidget(view)
-
-            self._setup_table_view(view, data)
-            slot = TableSlot(tid, data, table_summary(data), view)
-            view.input_slot = slot
-            self._inputs[tid] = slot
-
-            self.tabs.setCurrentIndex(self.tabs.indexOf(view))
-
-            self._set_input_summary(slot)
-
-            if isinstance(slot.summary.len, concurrent.futures.Future):
-                def update(_):
-                    QMetaObject.invokeMethod(
-                        self, "_update_info", Qt.QueuedConnection)
-
-                slot.summary.len.add_done_callback(update)
-
-        elif tid in self._inputs:
-            slot = self._inputs.pop(tid)
-            view = slot.view
-            view.hide()
-            view.deleteLater()
-            self.tabs.removeTab(self.tabs.indexOf(view))
-
-            current = self.tabs.currentWidget()
-            if current is not None:
-                self._set_input_summary(current.input_slot)
-        else:
-            self._set_input_summary(None)
+        current = self.tabs.currentWidget()
+        if current is not None:
+            self._set_input_summary(current.input_slot)
 
+    def handleNewSignals(self):
+        super().handleNewSignals()
         self.tabs.tabBar().setVisible(self.tabs.count() > 1)
+        data: Optional[Table] = None
+        current = self.tabs.currentWidget()
+        slot = None
+        if current is not None:
+            data = current.dataset
+            slot = current.input_slot
 
-        if data and self.__pending_selected_rows is not None:
+        if slot and isinstance(slot.summary.len, concurrent.futures.Future):
+            def update(_):
+                QMetaObject.invokeMethod(
+                    self, "_update_info", Qt.QueuedConnection)
+            slot.summary.len.add_done_callback(update)
+        self._set_input_summary(slot)
+
+        if data is not None and self.__pending_selected_rows is not None:
             self.selected_rows = self.__pending_selected_rows
             self.__pending_selected_rows = None
         else:
@@ -336,17 +351,12 @@ class OWDataTable(OWWidget):
             self.selected_cols = []
 
         self.set_selection()
-        self.unconditional_commit()
+        self.commit.now()
 
     def _setup_table_view(self, view, data):
         """Setup the `view` (QTableView) with `data` (Orange.data.Table)
         """
-        if data is None:
-            view.setModel(None)
-            return
-
         datamodel = RichTableModel(data)
-
         rowcount = data.approx_len()
 
         if self.color_by_class and data.domain.has_discrete_class:
@@ -356,11 +366,11 @@ class OWDataTable(OWWidget):
             color_schema = None
         if self.show_distributions:
             view.setItemDelegate(
-                gui.TableBarItem(
-                    self, color=self.dist_color, color_schema=color_schema)
+                TableBarItemDelegate(
+                    view, color=self.dist_color, color_schema=color_schema)
             )
         else:
-            view.setItemDelegate(QStyledItemDelegate(self))
+            view.setItemDelegate(TableDataDelegate(view))
 
         # Enable/disable view sorting based on data's type
         view.setSortingEnabled(is_sortable(data))
@@ -461,7 +471,7 @@ class OWDataTable(OWWidget):
                 s = btn.style().sizeFromContents(
                     QStyle.CT_HeaderSection,
                     opt, QSize(),
-                    btn).expandedTo(QApplication.globalStrut())
+                    btn)
                 if s.isValid():
                     table.verticalHeader().setMinimumWidth(s.width())
             except Exception:
@@ -543,15 +553,10 @@ class OWDataTable(OWWidget):
                     + format_part(summary.M))
         return text
 
-    def _set_output_summary(self, output):
-        summary = len(output) if output else self.info.NoOutput
-        details = format_summary_details(output) if output else ""
-        self.info.set_output_summary(summary, details)
-
     def _on_select_all(self, _):
         data_info = self.tabs.currentWidget().input_slot.summary
         if len(self.selected_rows) == data_info.len \
-                and len(self.selected_cols) == len(data_info.domain):
+                and len(self.selected_cols) == len(data_info.domain.variables):
             self.tabs.currentWidget().clearSelection()
         else:
             self.tabs.currentWidget().selectAll()
@@ -588,7 +593,7 @@ class OWDataTable(OWWidget):
 
     def _on_show_variable_labels_changed(self):
         """The variable labels (var.attribues) visibility was changed."""
-        for slot in self._inputs.values():
+        for slot in self._inputs:
             self._update_variable_labels(slot.view)
 
     def _on_distribution_color_changed(self):
@@ -604,17 +609,17 @@ class OWDataTable(OWWidget):
             else:
                 color_schema = None
             if self.show_distributions:
-                delegate = gui.TableBarItem(self, color=self.dist_color,
-                                            color_schema=color_schema)
+                delegate = TableBarItemDelegate(widget, color=self.dist_color,
+                                                color_schema=color_schema)
             else:
-                delegate = QStyledItemDelegate(self)
+                delegate = TableDataDelegate(widget)
             widget.setItemDelegate(delegate)
         tab = self.tabs.currentWidget()
         if tab:
             tab.reset()
 
     def _on_select_rows_changed(self):
-        for slot in self._inputs.values():
+        for slot in self._inputs:
             selection_model = slot.view.selectionModel()
             selection_model.setSelectBlocks(not self.select_rows)
             if self.select_rows:
@@ -640,7 +645,7 @@ class OWDataTable(OWWidget):
             self._set_input_summary(current.input_slot)
 
     def update_selection(self, *_):
-        self.commit()
+        self.commit.deferred()
 
     def set_selection(self):
         if self.selected_rows and self.selected_cols:
@@ -698,6 +703,7 @@ class OWDataTable(OWWidget):
             model = model.sourceModel()
         return model
 
+    @gui.deferred
     def commit(self):
         """
         Commit/send the current selected row/column selection.
@@ -711,7 +717,6 @@ class OWDataTable(OWWidget):
             # Selections of individual instances are not implemented
             # for SqlTables
             if isinstance(table, SqlTable):
-                self._set_output_summary(selected_data)
                 self.Outputs.selected_data.send(selected_data)
                 self.Outputs.annotated_data.send(None)
                 return
@@ -721,7 +726,7 @@ class OWDataTable(OWWidget):
 
             domain = table.domain
 
-            if len(colsel) < len(domain) + len(domain.metas):
+            if len(colsel) < len(domain.variables) + len(domain.metas):
                 # only a subset of the columns is selected
                 allvars = domain.class_vars + domain.metas + domain.attributes
                 columns = [(c, model.headerData(c, Qt.Horizontal,
@@ -747,7 +752,6 @@ class OWDataTable(OWWidget):
             else:
                 selected_data = table.from_table(domain, table, rowsel)
 
-        self._set_output_summary(selected_data)
         self.Outputs.selected_data.send(selected_data)
         self.Outputs.annotated_data.send(create_annotated_table(table, rowsel))
 
@@ -851,6 +855,8 @@ def is_sortable(table):
 
 if __name__ == "__main__":  # pragma: no cover
     WidgetPreview(OWDataTable).run(
-        [(Table("iris"), "iris"),
-         (Table("brown-selected"), "brown-selected"),
-         (Table("housing"), "housing")])
+        insert_dataset=[
+            (0, Table("iris")),
+            (1, Table("brown-selected")),
+            (2, Table("housing"))
+        ])
